// MuehleMain.java

/**
 * Changelog:
 * V 1.1 ----------------------------------
 * fixed game reset, loser may now challenge winner.
 * fixed beginning player when game is reset: loser begins.
 * game will now terminate, if relay server can't be reached.
 * added new win animation
 * optimized getValidMoveLocation (less code now)
 * optimized checkForMills
 * added information how many stones are left to place
 * Players can now win through blocking all of the opponents movements
 *
 * various messages fixed
 *
 * V1.1a -----------------------------------
 * flag isMouseEnabled instead of setMouseEnabled() due to buffered mouse event
 * message when starting drag revisted
 */

import java.awt.Color;
import java.util.ArrayList;
import java.util.Arrays;
import ch.aplu.tcp.*;
import javax.swing.JOptionPane;

import ch.aplu.jgamegrid.*;

public class MuhleMain extends GameGrid implements GGMouseListener
{
  private interface Command
  {
    int LCLICK = 0;
    int LRELEASE = 1;
    int LDRAG = 2;
    int YOU_MOVE = 3;
  }

  private enum PlayingStage
  {
    placing, moving, flying, removing, finished
  }

  private int[] nrStonesSet = new int[2];
  private final int timeout = 10000; // timeout for connection to relay
  private int currentPlayer;
  private ArrayList<String> badMoves = new ArrayList<String>();
  private PlayStone dragStone = null;
  private Location initialDragStoneLoc;
  private PlayingStage[] playingStage = new PlayingStage[2];
  private PlayingStage previousStage;
  private static boolean isClient;
  private final int nbStones = 9;
  private final String version = "1.1a";
  private String sessionID = "MillGame: ";
  private final String bridgeName = "MillServer";
  private final String agentName = "MillClient";
  private TcpAgent agent;
  private TcpBridge bridge;
  private volatile boolean isMouseEnabled = true;

  public MuhleMain()
  {
    super(7, 7, 70, null, "sprites/board.png", false);
    setBgColor(Color.white);
    setTitle("Nine Men's Morris (Mill) V " + version + " - Developed by Stefan Moser");
    setSimulationPeriod(200);

    String[] s = new String[7];
    s[0] = "_xx_xx_";
    s[1] = "x_x_x_x";
    s[2] = "xx___xx";
    s[3] = "___x___";
    s[4] = s[2];
    s[5] = s[1];
    s[6] = s[0];
    stringToBadMoves(s);

    String id = requestEntry("Enter a room name (>2 chars):");
    sessionID = sessionID + id;
    agent = new TcpAgent(sessionID, bridgeName);
    bridge = new TcpBridge(sessionID, bridgeName);
    addStatusBar(30);
    setStatusText("Connectiong to relay...");
    if (!connect())
    {
      terminate("Connection to relay failed. Terminating game now...");
    }
    if (isClient)
    {
      agent.addTcpAgentListener(new TcpAgentAdapter()
      {
        public void dataReceived(String source, int[] data)
        {
          execute(data);
        }

        public void notifyBridgeConnection(boolean connected)
        {
          if (!connected)
            terminate("Partner disconnected. Terminating game now...");
        }
      });
    }
    else // server
    {
      setStatusText("Connection established. Waiting for partner...");
      bridge.addTcpBridgeListener(new TcpBridgeAdapter()
      {
        public void pipeRequest(String source, String destination, int[] data)
        {
          execute(data);
        }

        public void notifyAgentConnection(String agent, boolean connected)
        {
          if (!connected)
            terminate("Partner disconnected. Terminating game now...");
        }
      });
    }
    init();
    show();
    if (!isClient)
      isMouseEnabled = false;  // Partner starts
    addMouseListener(this, GGMouse.lClick | GGMouse.lDrag | GGMouse.lRelease);
  }

  private void terminate(String text)
  {
    setStatusText(text);
    TcpTools.delay(4000);
    System.exit(0);
  }

  private boolean connect()
  {
    // First try to connect as bridge
    ArrayList<String> connectList = bridge.connectToRelay(timeout);
    if (connectList.size() >= 3)
    { //exit if there are already 2 players present
      terminate("Game in progress. Select another room.");
      //TODO: spectator mode?
    }
    if (connectList.isEmpty())
      return false; // Connection to relay failed
    if (!connectList.get(0).equals(bridgeName)) // bridge already connected
    {
      bridge.disconnect();
      connectList = agent.connectToRelay(agentName, timeout); // connect as agent
      if (connectList.isEmpty())
        return false;
      isClient = true;  // Client
    }
    else
      isClient = false;  // Server
    return true;
  }

  private void execute(int[] data)
  {
    GGMouse mouse = null;
    switch (data[0])
    {
      case Command.LCLICK:
        mouse = GGMouse.create(null, GGMouse.lClick, data[1], data[2]);
        mouseAction(mouse, false);
        break;
      case Command.LRELEASE:
        mouse = GGMouse.create(null, GGMouse.lRelease, data[1], data[2]);
        mouseAction(mouse, false);
        break;
      case Command.LDRAG:
        mouse = GGMouse.create(null, GGMouse.lDrag, data[1], data[2]);
        mouseAction(mouse, false);
        break;
      case Command.YOU_MOVE:
        if (!(playingStage[currentPlayer] == PlayingStage.finished))
        {
          if (nrStonesSet[currentPlayer] == nbStones)
        	  setStatusTextBW("Move a stone by dragging it with the mouse.");
          else
       	    setStatusTextBW("It's you to play!");
        }
        isMouseEnabled = true;
        activate();
        break;
    }
  }

  private void sendCommand(int command, int... data)
  {
    if (isClient)
      agent.sendCommand("", command, data);
    else
      bridge.sendCommand("", agentName, command, data);
  }

  public boolean mouseEvent(GGMouse mouse)
  {
    if (!isMouseEnabled)
      return true;
    return mouseAction(mouse, true);
  }

  private boolean mouseAction(GGMouse mouse, boolean mustSend)
  // mustSend true, if local mouse action; false, if mouse action reported 
  // from remote
  {
    Location mouseLoc = toLocation(mouse.getX(), mouse.getY());
    if (mustSend)
    {
      int command = -1;
      switch (mouse.getEvent())
      {
        case GGMouse.lClick:
          command = Command.LCLICK;
          break;
        case GGMouse.lRelease:
          command = Command.LRELEASE;
          break;
        case GGMouse.lDrag:
          //to prevent moving opponents stones on remote screen
          if (dragStone != null) // only send if there is a valid stone
            command = Command.LDRAG;
          break;
      }
      System.out.println("Drag: " + mouse.getX());
      sendCommand(command, mouse.getX(), mouse.getY());
    }

    if (mouse.getEvent() == GGMouse.lClick)
    {
      if (playingStage[currentPlayer] == PlayingStage.finished)
      {
        reset(mustSend); //reset game by clicking
        return true;
      }

      /**if current player is in removing stage and clicked on a
       * an actor -> check i
       * usually 0f it's an opponents actor & remove it then
       */
      if (playingStage[currentPlayer] == PlayingStage.removing)
      {
        PlayStone toBeRemovedPS = getPSat(mouseLoc);
        // don't do anything if it's own stone or unoccupied location
        if (toBeRemovedPS == null
          || toBeRemovedPS.belongsToPlayer() == currentPlayer)
          return true;

        if (checkForMills(mouseLoc)
          && enemyHasUnmilledStonesLeft(currentPlayer))
        {
          setStatusTextBW("Cannot remove from a closed mill!");
          return true;
        }

        removeActor(toBeRemovedPS);
        playingStage[currentPlayer] = previousStage;
        changePlayer(mustSend); // change player
        
		/* swap the other player to flying Stage
		 * if he's not in placing stage
         * and has 3 or fewer stones.
         */
        if (playingStage[currentPlayer] != PlayingStage.placing
          && getNrOfStones(currentPlayer) <= 3)
          playingStage[currentPlayer] = PlayingStage.flying;
        
        if (playingStage[currentPlayer] != PlayingStage.placing // currentPlayer at this point is remote!
          && getNrOfStones(currentPlayer) <= 2) // did local win?
        {
        	activateWinAnimation((currentPlayer + 1) % 2);
        	if (mustSend)
        		setStatusTextBW("You win! Wait for losers (possible) challenge.");
        	else setStatusTextBW("You Lost! Click to play another game.");
        }
        
        refresh();
        return true;
      }

      if (playingStage[currentPlayer] == PlayingStage.moving)
      { // at max 9 stones each
        if (mustSend)
          setStatusTextBW("Move a stone by dragging it with the mouse.");
        return true;
      }

      if (playingStage[currentPlayer] == PlayingStage.placing)
      {
        if (getOneActorAt(mouseLoc) != null)
        {
          if (mustSend) setStatusTextBW("This position is already occupied.");
          return true;
        }

        if (badMoves.contains(mouseLoc.toString()))
        {
        	if (mustSend) setStatusTextBW("Invalid move. Choose another location!");
          return true;
        }

        Actor playStone = new PlayStone(currentPlayer);
        addActor(playStone, mouseLoc);
        nrStonesSet[currentPlayer]++;
        if (nrStonesSet[currentPlayer] == nbStones)
          playingStage[currentPlayer] = PlayingStage.moving; // at max 9 stones each

        if (checkForMills(mouseLoc))
        {
          if (mustSend)
        	  setStatusTextBW("You may remove an enemy stone!");
          else setStatusTextBW("One of your stones is to be removed!");
          this.previousStage = playingStage[currentPlayer];
          playingStage[currentPlayer] = PlayingStage.removing;
          return true;
        }
        changePlayer(mustSend);
      }
    }

    /**
     * Dragging = either jumping or moving
     */
    if (mouse.getEvent() == GGMouse.lDrag)
    {
      if (!(playingStage[currentPlayer] == PlayingStage.moving
        || playingStage[currentPlayer] == PlayingStage.flying))
        return true;
      // is the same in moving and flying stage:
      if (playingStage[currentPlayer] == PlayingStage.moving
        || playingStage[currentPlayer] == PlayingStage.flying)
      {
        if (getOneActorAt(mouseLoc) == null && dragStone == null)
        {
          if (mustSend) setStatusTextBW("Drag a stone with your mouse");
          return true;
        }

        if (dragStone == null)
        {
          dragStone = getPSat(mouseLoc);
          initialDragStoneLoc = mouseLoc;
          if (mustSend) setStatusTextBW("Place stone by releasing the button.");
          if (dragStone.belongsToPlayer() != currentPlayer)
          {
        	if (mustSend) setStatusTextBW("You can only move your own stones!");
            dragStone = null; //reset dragStone
            return true;
          }
        }
        else
        {
          dragStone.setLocation(mouseLoc);
          this.refresh(); // refresh to show change
        }
      }
    }

    if (mouse.getEvent() == GGMouse.lRelease)
    {
      System.out.println("Registered release: " + (dragStone != null));
      // if the player is at flying stage:
      if (playingStage[currentPlayer] == PlayingStage.flying
        && dragStone != null)
      {
        if (isValidFlyingLocation(mouseLoc))
        {

          System.out.println("Release : " + dragStone);
          dragStone.setLocation(mouseLoc);
          dragStone = null;

          if (checkForMills(mouseLoc))
          {
        	  if (mustSend)
            	  setStatusTextBW("You may remove an enemy stone!");
              else setStatusTextBW("One of your stones is to be removed!");
            this.previousStage = playingStage[currentPlayer];
            playingStage[currentPlayer] = PlayingStage.removing;
          }
          else
            changePlayer(mustSend); //end move
        }
        else
        {
          dragStone.setLocation(this.initialDragStoneLoc);
          setStatusTextBW("That was an illegal move! Try again!");
          dragStone = null;
        }
      }

      // if the player is at moving stage:
      if (playingStage[currentPlayer] == PlayingStage.moving
        && dragStone != null)
      {
        if (isValidFlyingLocation(mouseLoc) //adds more conditions for moving (only adjacent!)
          && isValidMovingLocation(mouseLoc))
        {
          dragStone.setLocation(mouseLoc);
          dragStone = null;

          if (checkForMills(mouseLoc))
          {
        	if (mustSend)
              setStatusTextBW("You may remove an enemy stone!");
            else setStatusTextBW("One of your stones is to be removed!");
            this.previousStage = playingStage[currentPlayer];
            playingStage[currentPlayer] = PlayingStage.removing;
          }
          else
            changePlayer(mustSend); // end move
        }
        else
        {
          dragStone.setLocation(this.initialDragStoneLoc);
          if (mustSend) setStatusTextBW("That was an illegal move! Try again!");
          dragStone = null;
        }
      }
     
    }
    this.refresh();
    return true;
  }

  private void activateWinAnimation(int winner) {
	  ArrayList<Actor> playStones = getActors();
      for (Actor a : playStones)
      {
        PlayStone pS = (PlayStone)a;
        if (pS.belongsToPlayer() == winner) {
        	pS.show(2);
        	pS.setActEnabled(true);
        }
        else pS.setActEnabled(false);
      }
      doRun();
      Arrays.fill(playingStage, PlayingStage.finished);
}

/**
   * Returns PlayStone at indicated location.
   */
  private PlayStone getPSat(Location loc)
  {
    return (PlayStone)getOneActorAt(loc);
  }
  
  /**
   * Checks every single stone of the parameter "player" if they're in a mill or not.
   * If all of them are milled, false is returned. If not, true is returned.
   */
  private boolean enemyHasUnmilledStonesLeft(int player)
  {
    int enemyPlayer = (player + 1) % 2;
    int nrOfStonesInMill = 0;
    ArrayList<Actor> allPlayStones = this.getActors();
    for (Actor a : allPlayStones)
    {
      PlayStone pS = (PlayStone)a;
      if (pS.belongsToPlayer() == enemyPlayer
        && checkForMills(pS.getLocation()))
        nrOfStonesInMill++;
    }
    System.out.println("" + nrOfStonesInMill + getNrOfStones(enemyPlayer)); //debug
    return (nrOfStonesInMill < getNrOfStones(enemyPlayer));
  }

  /**
   * Checks for mills with a new playStone placed at
   * the indicated location. (Middle row and column have to
   * be treated differently (see helper-methods).
   *
   * Returns true if a mill was completed
   */
  private boolean checkForMills(Location loc)
  {
    return (checkForMillsHorizontally(loc)
      || checkForMillsVertically(loc));
  }

  private boolean checkForMillsHorizontally(Location loc)
  {
    int stoneCounter = 0;
    int stopCheckingPoint = this.getNbHorzCells(); //usually check all lines
    int startCheckingPoint = 0;
    if (loc.y == 3)
    {
      if (loc.x < 3)
      { //only check first 3
        startCheckingPoint = 0;
        stopCheckingPoint = 3;
      }
      if (loc.x > 3)
      { //or last 3
        startCheckingPoint = this.getNbHorzCells() - 3;
        stopCheckingPoint = this.getNbHorzCells();
      }
    }
    for (int i = startCheckingPoint; i < stopCheckingPoint; i++)
    {
      Location checkLoc = new Location(i, loc.y);
      if (getPSat(checkLoc) != null
    		  && getPSat(checkLoc).belongsToPlayer() == getPSat(loc).belongsToPlayer())
      {
        stoneCounter++;
      }
    }
    return (stoneCounter >= 3);
  }

  private boolean checkForMillsVertically(Location loc)
  {
    int stoneCounter = 0;
    int startCheckingPoint = 0;
    int stopCheckingPoint = this.getNbVertCells(); //usually check all lines
    if (loc.x == 3)
    {
      if (loc.y < 3)
      { //only check first 3
        startCheckingPoint = 0;
        stopCheckingPoint = 3;
      }
      if (loc.y > 3)
      { //or last 3
        startCheckingPoint = this.getNbVertCells() - 3;
        stopCheckingPoint = this.getNbVertCells();
      }
    }
    for (int i = startCheckingPoint; i < stopCheckingPoint; i++)
    {
      Location checkLoc = new Location(loc.x, i);
      if (getPSat(checkLoc) != null
    	        && getPSat(checkLoc).belongsToPlayer() == getPSat(loc).belongsToPlayer())
      {
        stoneCounter++;
      }
    }
    return (stoneCounter >= 3);
  }

  /**
   * Checks if all possible movements for every stone of "player"
   * @return true if all movements are blocked
   */
  private boolean areAllMovementsBlocked(int player) {
	  //only applies to movement stage -> return false if other stage
	  if (playingStage[player] != PlayingStage.moving) 
		  return false;
	  
	  ArrayList<PlayStone> allStonesOfPlayer = new ArrayList<PlayStone>();
	  for (Actor a: getActors(PlayStone.class))
		  if (((PlayStone)a).belongsToPlayer() == player)
			  allStonesOfPlayer.add((PlayStone)a);
			  
	  System.out.println(allStonesOfPlayer);
	  
	  //Checks all neighbouring fields (distance 3) on their validity as moving-position
	  for (PlayStone p: allStonesOfPlayer) {
		  for (Location checkLoc: p.getLocation().getNeighbourLocations(3)) //check all neighbour locations
			  if (getPSat(checkLoc) == null) { //has to be unoccupied
				  initialDragStoneLoc = p.getLocation();
				  if (isInGrid(checkLoc) && isValidFlyingLocation(checkLoc)
						  && isValidMovingLocation(checkLoc)) // and valid
					  {
					  System.out.println("returned false bc of: " + checkLoc); //debug
					  return false;
					  }
			  }
				  
	  }
	  return true;
  }

  /**
   * Returns true if the submitted Location is a valid
   * flying Location
   * 
   */
  private boolean isValidFlyingLocation(Location loc)
  {
    return (!badMoves.contains(loc.toString())
      && !loc.toString().equals(initialDragStoneLoc.toString())
      && getActorsAt(loc).size() < 2);
  }

  /**
   * Returns true if the location is a valid move location for a
   * stone originating in "initialDragLocation"
   * (Defines a maximum distance in each direction for every row/column)
   */
  
  private boolean isValidMovingLocation(Location loc)
  {
    int maxDistanceX = getMaxMoveDistance(loc.y); //symmetrical field
    int maxDistanceY = getMaxMoveDistance(loc.x); // -> same method can be used

    return ((loc.x == initialDragStoneLoc.x || loc.y == initialDragStoneLoc.y)
      && Math.abs(initialDragStoneLoc.x - loc.x) <= maxDistanceX
      && Math.abs(initialDragStoneLoc.y - loc.y) <= maxDistanceY);
  }
  
  private int getMaxMoveDistance(int pos) {
	  switch (pos)
	    {
	      case 0:
	        return 3;
	      case 1:
	        return 2;
	      case 5:
	        return 2;
	      case 6:
	        return 3;
	      default:
	        return 1;  //for columns/rows 2, 3 and 4
	    }
  }

  /**
   * Adds "Black player: " respectively " White player: "
   * to the message and sets it as title.
   * @param msg
   */
  private void setStatusTextBW(String msg)
  {
	int localPlayer;
	String stonesLeft = "";
	if (isClient)
		localPlayer = 0;
	else localPlayer = 1;
	if (nrStonesSet[localPlayer] < nbStones)
			stonesLeft = " " + 
			(nbStones-nrStonesSet[localPlayer]) +
			" stones left to place.";
	
    if (isClient)
      setStatusText("<WHITE STONES> " + msg + stonesLeft);
    else
      setStatusText("<BLACK STONES> " + msg + stonesLeft);
  }

  private int getNrOfStones(int player)
  {
    int nrOfStones = 0;
    ArrayList<Actor> allPlayStones = this.getActors();
    for (Actor a : allPlayStones)
    {
      PlayStone pS = (PlayStone)a;
      if (pS.belongsToPlayer() == player)
        nrOfStones++;
    }
    return nrOfStones;
  }

  /**
   * Changes currentPlayer and states it in the title.
   */
  private void changePlayer(boolean mustSend)
  {
    isMouseEnabled = false;
    currentPlayer = (currentPlayer + 1) % 2;
    //TODO: working here
    //every term ends here! -> Check here for blocked!
    setStatusTextBW("Wait. It's the partner to play!");
	if (areAllMovementsBlocked(currentPlayer)) {
		{
        	activateWinAnimation((currentPlayer + 1) % 2);
        	if (mustSend)
        		setStatusTextBW("You win! Wait for losers (possible) challenge.");
        	else setStatusTextBW("You Lost! Click to play another game.");
        }
	}
    if (!mustSend)
      return;
    sendCommand(Command.YOU_MOVE);
  }

  /**
   * Adds all invalid locations (as Strings) to "BadMoves"
   */
  private void stringToBadMoves(String[] s)
  {
    for (int row = 0; row < s.length; row++)
    {
      int column = 7;
      while (column >= 0)
      {
        column = s[row].lastIndexOf('x', column - 1);
        if (column != -1)
        { // -1 is returned if not present
          badMoves.add((new Location(column, row)).toString());
        }
      }
    }
  }

  public void reset(boolean mustSend)
  {
    doPause();
    removeAllActors();
    refresh();
    Arrays.fill(playingStage, PlayingStage.placing);
    Arrays.fill(nrStonesSet, 0);

    if (mustSend) {
    	setStatusTextBW("Game reset. It's you to set the first stone!");
    }
    else {
        setStatusTextBW("Game reset. Wait for partner...");
        isMouseEnabled = false;
    }
  }

  /**
   * Initializes the game
   */
  public void init()
  {
    Arrays.fill(playingStage, PlayingStage.placing);
    Arrays.fill(nrStonesSet, 0);
    if (isClient)
      setStatusTextBW("Partner connected. It's you to set the first stone!");
    else
      setStatusTextBW("Connection established. Wait for partner...");
    currentPlayer = 0; //white starts
  }

  private String requestEntry(String prompt)
  {
    String entry = "";
    while (entry.length() < 3)
    {
      entry = JOptionPane.showInputDialog(null, prompt, "");
      if (entry == null)
        System.exit(0);
    }
    return entry.trim();
  }

  public static void main(String[] args)
  {
    new MuhleMain();
  }
}
